{% macro item_desc(b, project_url, project_name) -%}
  <p>
    Date: {{ b.date }}<br/>
    Project home: <a href="{{ project_url }}">{{ project_name|e }}</a><br/>
    Builder summary: <a href="{{ b.summary_link }}">{{ b.name }}</a><br/>
    Build details: <a href="{{ b.link }}">Build {{ b.number }}</a><br/>
    Author list: <b>{{ b.responsible_users|join(', ') }}</b><br/>
    Failed step(s): <b>{{ b.failed_steps|join(', ') }}</b><br/>
  </p>
  <p>
    <i>Last lines of the build log:</i>
  </p>
{%- endmacro %}
