{% extends "layout.html" %}
{% from "box_macros.html" import box %}

{% block content %}
<table>
  {% for r in rows %}
    <tr>{{ box(r) }}</tr>
  {% endfor %}

  <tr>{{ box(build) }}</tr>
</table>
{% endblock %}
