{% from 'build_line.html' import build_table, build_line %}
{% import 'forms.html' as forms %}

{% extends "layout.html" %}
{% block content %}
<h1>Buildslave: {{ slavename|e }}</h1>

<div class="column">

{% if current %}
  <h2>Currently building:</h2>
  <ul>
  {% for b in current %}
   <li>{{ build_line(b, True) }}
     <form method="post" action="{{ b.buildurl }}/stop" class="command stopbuild" style="display:inline">
       <input type="submit" value="Stop Build" />
       <input type="hidden" name="url" value="{{ this_url }}" />
     </form>
   </li>
  {% endfor %}
  </ul>
{% else %}
  <h2>No current builds</h2>
{% endif %}

<h2>Recent builds</h2>
{{ build_table(recent, True) }}
  
</div>
<div class="column">
{% if access_uri %}
  <a href="{{ access_uri|e }}">Click to Access Slave</a>
{% endif %}

{% if admin %}
  <h2>Administrator</h2>
  <p>{{ admin|email }}</p>
{% endif %}

{% if host %}
  <h2>Slave information</h2>
  <pre>{{ host|e }}</pre>
{% endif %}

{% if not slave.isConnected() %}
  <p>(not connected)</p>
{% elif authz.advertiseAction('gracefulShutdown') %}
  <h2>Graceful Shutdown</h2>
  {% if slave.getGraceful() %}
    <p>Slave will shut down gracefully when it is idle.</p>
  {% else %}
    {{ forms.graceful_shutdown(shutdown_url, authz) }}
  {% endif %}
{% endif %}
</div>
  
{% endblock %}
